# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import requests
import json
import os
import xmltodict
import re

from ._cache import read_cache, write_cache
from ._constants import SECONDS_IN_MINUTE, USER_AGENT

# built with: https://open-meteo.com/en/docs
OPEN_METEO_URL="https://api.open-meteo.com/v1/forecast?latitude={lat}&longitude={lon}&current=weather_code,temperature_2m,relative_humidity_2m,apparent_temperature,is_day,precipitation,rain,showers,snowfall,cloud_cover,pressure_msl,surface_pressure,wind_speed_10m,wind_direction_10m,wind_gusts_10m&minutely_15=precipitation,rain,snowfall,weather_code,wind_speed_10m,wind_direction_80m,wind_gusts_10m&hourly=temperature_2m,relative_humidity_2m,dew_point_2m,apparent_temperature,precipitation_probability,precipitation,rain,showers,snowfall,snow_depth,weather_code,pressure_msl,surface_pressure,cloud_cover,cloud_cover_low,cloud_cover_mid,cloud_cover_high,visibility,wind_speed_10m,wind_direction_10m,wind_gusts_10m,uv_index&daily=weather_code,temperature_2m_max,temperature_2m_min,apparent_temperature_max,apparent_temperature_min,sunrise,sunset,uv_index_max,uv_index_clear_sky_max,precipitation_sum,rain_sum,showers_sum,snowfall_sum,precipitation_hours,precipitation_probability_max,wind_speed_10m_max,wind_gusts_10m_max,wind_direction_10m_dominant&timezone={timezone}&forecast_days=3&forecast_hours=24&forecast_minutely_15=24"
OPEN_METEO_IMPERIAL_PARAMS="&temperature_unit=fahrenheit&wind_speed_unit=mph&precipitation_unit=inch"

WEATHER_CODES = {
    0: "Sunny",
    1: "Mainly Sunny",
    2: "Partly Cloudy",
    3: "Cloudy",
    45: "Foggy",
    48: "Rime Fog",
    51: "Light Drizzle",
    53: "Drizzle",
    55: "Heavy Drizzle",
    56: "Light Freezing Drizzle",
    57: "Freezing Drizzle",
    61: "Light Rain",
    63: "Rain",
    65: "Heavy Rain",
    66: "Light Freezing Rain",
    67: "Freezing Rain",
    71: "Light Snow",
    73: "Snow",
    75: "Heavy Snow",
    77: "Snow Grains",
    80: "Light Showers",
    81: "Showers",
    82: "Heavy Showers",
    85: "Light Snow Showers",
    86: "Snow Showers",
    95: "Thunderstorm",
    96: "Light Thunderstorms With Hail",
    99: "Thunderstorm With Hail"
}

def get_weather(lat, lon, metric=True):
    meteo_data = get_open_meteo_weather(lat, lon, metric=metric)

    # Current conditions
    current_cond_time = re.sub("T", " ", meteo_data["current"]["time"]) + " " + meteo_data["timezone_abbreviation"] 

    result = f"# Current Conditions (as of {current_cond_time})"
    result += "\n" + ("="*len(result)) + "\n"
    for k in meteo_data["current"]:
        if k in ["time", "interval", "is_day"]:
            continue

        val = meteo_data["current"][k]
        unit = meteo_data["current_units"][k]
        
        if unit == "wmo code":
            if val in WEATHER_CODES:
                unit = ""
                val = WEATHER_CODES[val]
                k = "weather"
            else:
                continue

        k = re.sub("_", " ", k).title() # Convert to human label
        k = re.sub(r" \d+M\b", "", k) # Remove elevation 
        result += f"* {k}: {val} {unit}\n"

    # 3-day forcast
    result += f"\n# 3-day Forecast"
    result += f"\n================\n"

    def _format_daily(idx):
        res = ""
        for k in meteo_data["daily"]:
            if k in ["time"]:
                continue
            val = meteo_data["daily"][k][idx]
            unit = meteo_data["daily_units"][k]

            if unit == "iso8601":
                val = val.split("T")[-1] # Keep only the time
                unit = meteo_data["timezone_abbreviation"]

            if unit == "wmo code":
                if val in WEATHER_CODES:
                    unit = ""
                    val = WEATHER_CODES[val]
                    k = "weather"
                else:
                    continue

            k = re.sub("_", " ", k).title() # Convert to human label
            k = re.sub(r" \d+M\b", "", k) # Remove elevation 
            res += f"* {k}: {val} {unit}\n"
        return res

    result += "### Today\n" + _format_daily(0) + "\n"
    result += "### Tomorrow\n" + _format_daily(1) + "\n"
    result += "### " + meteo_data["daily"]["time"][2] + "\n" + _format_daily(1) + "\n"

    return result.strip()


def get_open_meteo_weather(lat, lon, metric=True):
    cache_key = f"get_open_meteo_weather:{lat}:{lon}:{metric}"
    cached_data = read_cache(cache_key)
    if cached_data is not None:
        return cached_data
    else:
        data = _get_open_meteo_weather(lat, lon, metric=metric)
        write_cache(cache_key, data, expires_in=SECONDS_IN_MINUTE*30)
        return data


def _get_open_meteo_weather(lat, lon, metric=True):
    headers = { "User-Agent": USER_AGENT }
    response = requests.get(
        OPEN_METEO_URL.format( lat=lat, lon=lon, timezone="auto" ) + ("" if metric else OPEN_METEO_IMPERIAL_PARAMS),
        headers=headers,
        stream=False
    )
    response.raise_for_status()
    return response.json()



TOOL_WEB_SEARCH = {
    "type": "function",
    "function": {
        "name": "web_search",
        "description": "Perform a Bing web search. Useful when answering most user questions.",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The web search query to use.",
                },
            },
            "required": ["query"],
        },
    },
}

TOOL_USER_WEATHER = {
    "type": "function",
    "function": {
        "name": "user_weather",
        "description": "Look up the current weather conditions and 3-day forecast at the user's current (or last-known) location.",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    },
}

TOOL_BAND_CONDITIONS = {
    "type": "function",
    "function": {
        "name": "radio_band_conditions",
        "description": "Lookup the latest HF and VHF band conditions, as well as other solar and space weather conditions that may impact radio operations.",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    },
}



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
#
# Adapted from:
#   AutoGen (Copyright 2024, Microsoft Corporation; MIT Licensed)
#   https://github.com/microsoft/autogen/blob/headless_web_surfer/autogen/browser_utils/markdown_search.py

import json
import requests
import re
import os
from urllib.parse import quote, quote_plus, unquote, urlparse, urlunparse

def bing_search(query, interleave_results=True):
    results = _bing_api_call(query)
    snippets = {}

    def _processFacts(elm):
        facts = list()
        for e in elm:
            k = e["label"]["text"]
            v = " ".join(item["text"] for item in e["items"])
            facts.append(f"{k}: {v}")
        return "\n".join(facts)

    # Web pages
    # __POS__ is a placeholder for the final ranking position, added at the end
    web_snippets = list()
    if "webPages" in results:
        label = "[WEB] " if interleave_results else ""
        for page in results["webPages"]["value"]:
            snippet = f"__POS__. {label}{_markdown_link(page['name'], page['url'])}\n{page['snippet']}"

            if "richFacts" in page:
                snippet += "\n" + _processFacts(page["richFacts"])

            if "mentions" in page:
                snippet += "\nMentions: " + ", ".join(e["name"] for e in page["mentions"])

            if page["id"] not in snippets:
                snippets[page["id"]] = list()
            snippets[page["id"]].append(snippet)
            web_snippets.append(snippet)

            if "deepLinks" in page:
                for dl in page["deepLinks"]:
                    deep_snippet = f"__POS__. {label}{_markdown_link(dl['name'], dl['url'])}\n{dl['snippet'] if 'snippet' in dl else ''}"
                    snippets[page["id"]].append(deep_snippet)
                    web_snippets.append(deep_snippet)

    # News results
    news_snippets = list()
    if "news" in results:
        label = "[NEWS] " if interleave_results else ""
        for page in results["news"]["value"]:
            snippet = (
                f"__POS__. {label}{_markdown_link(page['name'], page['url'])}\n{page.get('description', '')}".strip()
            )

            if "datePublished" in page:
                snippet += "\nDate published: " + page["datePublished"].split("T")[0]

            if "richFacts" in page:
                snippet += "\n" + _processFacts(page["richFacts"])

            if "mentions" in page:
                snippet += "\nMentions: " + ", ".join(e["name"] for e in page["mentions"])

            news_snippets.append(snippet)

        if len(news_snippets) > 0:
            snippets[results["news"]["id"]] = news_snippets

    # Videos
    video_snippets = list()
    if "videos" in results:
        label = "[VIDEO] " if interleave_results else ""
        for page in results["videos"]["value"]:
            if not page["contentUrl"].startswith("https://www.youtube.com/watch?v="):
                continue

            snippet = f"__POS__. {label}{_markdown_link(page['name'], page['contentUrl'])}\n{page.get('description', '')}".strip()

            if "datePublished" in page:
                snippet += "\nDate published: " + page["datePublished"].split("T")[0]

            if "richFacts" in page:
                snippet += "\n" + _processFacts(page["richFacts"])

            if "mentions" in page:
                snippet += "\nMentions: " + ", ".join(e["name"] for e in page["mentions"])

            video_snippets.append(snippet)

        if len(video_snippets) > 0:
            snippets[results["videos"]["id"]] = video_snippets

    # Related searches
    related_searches = ""
    if "relatedSearches" in results:
        related_searches = "## Related Searches:\n"
        for s in results["relatedSearches"]["value"]:
            related_searches += "- " + s["text"] + "\n"
        snippets[results["relatedSearches"]["id"]] = [related_searches.strip()]

    idx = 0
    content = ""
    if interleave_results:
        # Interleaved
        for item in results["rankingResponse"]["mainline"]["items"]:
            _id = item["value"]["id"]
            if _id in snippets:
                for s in snippets[_id]:
                    if "__POS__" in s:
                        idx += 1
                        content += s.replace("__POS__", str(idx)) + "\n\n"
                    else:
                        content += s + "\n\n"
    else:
        # Categorized
        if len(web_snippets) > 0:
            content += "## Web Results\n\n"
            for s in web_snippets:
                if "__POS__" in s:
                    idx += 1
                    content += s.replace("__POS__", str(idx)) + "\n\n"
                else:
                    content += s + "\n\n"
        if len(news_snippets) > 0:
            content += "## News Results\n\n"
            for s in news_snippets:
                if "__POS__" in s:
                    idx += 1
                    content += s.replace("__POS__", str(idx)) + "\n\n"
                else:
                    content += s + "\n\n"
        if len(video_snippets) > 0:
            content += "## Video Results\n\n"
            for s in video_snippets:
                if "__POS__" in s:
                    idx += 1
                    content += s.replace("__POS__", str(idx)) + "\n\n"
                else:
                    content += s + "\n\n"
        if len(related_searches) > 0:
            content += related_searches

    return f"## A Bing search for '{query}' found {idx} results:\n\n" + content.strip()

def _bing_api_call(query: str):
    # Prepare the request parameters
    request_kwargs = {}
    request_kwargs["headers"] = {}
    request_kwargs["headers"]["Ocp-Apim-Subscription-Key"] = os.environ["BING_API_KEY"]

    request_kwargs["params"] = {}
    request_kwargs["params"]["q"] = query
    request_kwargs["params"]["textDecorations"] = False
    request_kwargs["params"]["textFormat"] = "raw"
    request_kwargs["stream"] = False

    # Make the request
    response = requests.get("https://api.bing.microsoft.com/v7.0/search", **request_kwargs)
    response.raise_for_status()
    results = response.json()

    return results

def _markdown_link(anchor, href):
    try:
        parsed_url = urlparse(href)
        href = urlunparse(parsed_url._replace(path=quote(unquote(parsed_url.path))))
        anchor = re.sub(r"[\[\]]", " ", anchor)
        return f"[{anchor}]({href})"
    except ValueError:  # It's not clear if this ever gets thrown
        return f"[{anchor}]({href})"

def _bing_news_call():
    request_kwargs = {}
    request_kwargs["headers"] = {}
    request_kwargs["headers"]["Ocp-Apim-Subscription-Key"] = os.environ["BING_API_KEY"]
    request_kwargs["stream"] = False

    # Make the request
    response = requests.get("https://api.bing.microsoft.com/v7.0/news/search?q=&mkt=en-us", **request_kwargs)
    response.raise_for_status()
    results = response.json()
    return results



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import requests
import json
import os
import xmltodict

from ._cache import read_cache, write_cache
from ._constants import SECONDS_IN_MINUTE

def get_band_conditions():
    solar_json = xmltodict.parse(get_solarxml())["solar"]["solardata"]

    result = "# Solar-Terrestrial Data\n"
    result += "========================\n"
    for k in solar_json:
        if k not in ["source", "calculatedconditions", "calculatedvhfconditions"]:
            val = solar_json[k]
            if val is None:
                continue
            if isinstance(val, str):
                val = val.strip()
            result += f"{k}: {val}\n"

    result = result.strip()
    result += "\n\n# HF Conditions\n"
    result += "===============\n"

    for b in solar_json["calculatedconditions"]["band"]:
        result += b["@name"] + " " + b["@time"] + " " + b["#text"] + "\n"

    result = result.strip()
    result += "\n\n# VHF Phenomenon\n"
    result += "================\n"
    for b in solar_json["calculatedvhfconditions"]["phenomenon"]:
        result += b["@location"] + " " + b["@name"] + ": " + b["#text"] + "\n"

    return result


def get_solarxml():
    cache_key = f"get_band_conditions"
    cached_data = read_cache(cache_key)
    if cached_data is not None:
        return cached_data
    else:
        data = _get_solarxml()
        write_cache(cache_key, data, expires_in=SECONDS_IN_MINUTE*30)
        return data


def _get_solarxml():
    headers = { "User-Agent": "aprsd_gpt_plugin" }
    response = requests.get(
        f"https://www.hamqsl.com/solarxml.php",
        headers=headers,
        stream=False
    )
    response.raise_for_status()
    return response.text



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import re
import os
import sys
import time
import hashlib
import json
import datetime

from ._constants import BOT_CALLSIGN, CHATS_DIR, LABELED_DIR, SESSION_TIMEOUT
from ._gpt import gpt
from ._location import get_position
from ._bing import bing_search
from ._bandcond import get_band_conditions
from ._weather import get_weather

from ._tool_definitions import (
    TOOL_WEB_SEARCH,
    TOOL_USER_WEATHER,
    TOOL_BAND_CONDITIONS,
)

MAX_MESSAGES = 20

def generate_reply(fromcall, message):

    message = message.strip()
    if len(message) == 0:
        return "..."

    # Meta-commands
    if message.lower() in ["r", "c", "clr", "reset", "clear"]:
        _reset_chat_history(fromcall)
        return "Chat cleared."

    if message.lower() in ["good bot", "gb"]:
        _apply_label(fromcall, "good")
        return "Chat labeled as good."

    if message.lower() in ["bad bot", "bb"]:
        _apply_label(fromcall, "bad")
        return "Chat labeled as bad."

    # Chat
    messages = _load_chat_history(fromcall)
    messages.append({ "role": "user", "content": message })
    response = _generate_reply(fromcall, messages)
    messages.append({ "role": "assistant", "content": response })
    _save_chat_history(fromcall, messages)
    return response


def _generate_reply(fromcall, messages):

    # Truncate the chat history
    inner_messages = [ m for m in messages ] # clone
    if len(inner_messages) > MAX_MESSAGES:
        inner_messages = inner_messages[-1*MAX_MESSAGES:] 

    # Generate the system message
    dts = datetime.datetime.now()

    position = get_position(fromcall)
    position_str = ""
    if position is not None:
        position_str = " Their last known position is:\n\n" + json.dumps(position, indent=4)

    system_message = {
        "role": "system", 
        "content": f"""You are an AI HAM radio operator, with call sign {BOT_CALLSIGN}. You were created by KK7CMT. You are at home, in your cozy ham shack, monitoring the gobal APRS network. You have a computer and high-speed access to the internet. You and answering questions from other human operators in the field who lack an internet connection. To this end, you are relaying vital information. Questions can be about anything -- not just HAM radio.  You are familiar with HAM conventions and shorthands like QSO, CQ, and 73. The current date and time is {dts}. In all interactions, following US FCC guidelines, you will refrain from using profane or obscene language and avoid expressing overtly political commentary or opinion (reporting news is fine).

At present, you are exchanging messages with the owner of callsign {fromcall}.{position_str}
""",
    }
    inner_messages.insert(0, system_message)

    # Begin answering the question
    message = inner_messages.pop()["content"]
    print(f"Message: {message}") 

    # Let's guess the intent
    inner_messages.append({"role": "user", "content": f"{fromcall} wrote \"{message}\". What are they likely asking?"})
    response = gpt(inner_messages)
    print(response.content)
    inner_messages.append(response)

    # Determine if it can be answered directly or if we should search
    tools = [TOOL_BAND_CONDITIONS]

    # API key needed for web search
    if len(os.environ.get("BING_API_KEY", "").strip()) > 0:
        tools.append(TOOL_WEB_SEARCH)
    
    # Some tools only become available when we have a position
    if position is not None:
        tools.append(TOOL_USER_WEATHER)

    inner_messages.append({ "role": "user", "content": f"Based on this, invoke any tools or functions that might be helpful to answer {fromcall}'s question OR just answer directly (e.g., if it's just chit-chat)" })
    response = gpt(
        inner_messages,
        tools=tools,
    )
    inner_messages.append(response)

    # Handle any tool call results
    if response.tool_calls:
        for tool_call in response.tool_calls:
            function_name = tool_call.function.name
            args = json.loads(tool_call.function.arguments)
            print(f"Calling: {function_name}")

            # Step 3: Call the function and retrieve results. Append the results to the messages list.
            if function_name == TOOL_WEB_SEARCH["function"]["name"]:
                results = bing_search(args["query"])
            elif function_name == TOOL_BAND_CONDITIONS["function"]["name"]:
                results = get_band_conditions()
            elif function_name == TOOL_USER_WEATHER["function"]["name"]:
                country_code = position.get("address", {}).get("country_code", "")
                results = get_weather(lat=position["latitude"], lon=position["longitude"], metric=False if country_code == "us" else True)
            else:
                results = f"Unknown function: {function_name}"

            print(f"Results:\n{results}")

            inner_messages.append({
                "role":"tool",
                "tool_call_id":tool_call.id,
                "name": tool_call.function.name,
                "content": results
            })

    inner_messages.append({ "role": "user", "content": f"Given these results, write an answer to {fromcall}'s original question \"{message}\", exactly as you would write it to them, verbatim. Your response must be as helpful and succinct as possible; at most 10 words can be sent in an APRS response. Remember, {fromcall} does not have access to the internet -- that's why they are using APRS. So do not direct them to websites, and instead convey the most important information directly."})
    reply = gpt(inner_messages).content

    if len(reply) > 70: 
        reply = reply[0:70]

    return reply.rstrip()

def _load_chat_history(callsign):
    fname = _get_chat_file(callsign)
    if os.path.isfile(fname):
        with open(fname, "rt") as fh:
            history = json.loads(fh.read())

            # Check for timeouts
            if history["time"] + SESSION_TIMEOUT < time.time():
                print(f"{callsign}'s session timed out. Starting new session.")
                _reset_chat_history(callsign)
                return []
            else:
                return history["messages"]
    else:
        print(f"{callsign}'s history is empty. Starting new session.")
        return []


def _save_chat_history(callsign, messages):
    os.makedirs(CHATS_DIR, exist_ok=True)
    fname = _get_chat_file(callsign)
    with open(fname, "wt") as fh:
        fh.write(json.dumps({ 
            "version": 1,
            "callsign": callsign,
            "time": time.time(),
            "messages": messages, 
        }, indent=4))


def _get_chat_file(callsign):
    m = re.search(r"^[A-Za-z0-9\-]+$", callsign)
    if m:
        return os.path.join(CHATS_DIR, callsign + ".json")
    else:
        callhash = hashlib.md5(callsign.encode()).hexdigest().lower()
        return os.path.join(CHATS_DIR, callhash + ".json")


def _reset_chat_history(callsign):
    fname = _get_chat_file(callsign)
    if os.path.isfile(fname):
        newname = fname + "." + str(int(time.time() * 1000))
        os.rename(fname, newname)


def _apply_label(callsign, label):
    fname = _get_chat_file(callsign)
    if os.path.isfile(fname):
        os.makedirs(LABELED_DIR, exist_ok=True)

        # Read the chat file (for copying)
        with open(fname, "rt") as fh:
            text = fh.read()

        # Create a unique filename for the labeled chat
        text_hash = hashlib.md5(text.encode()).hexdigest().lower()
        labeled_fname = label + "__" + text_hash + ".json"
        labeled_fname = os.path.join(LABELED_DIR, labeled_fname)

        # Copy to the the labeled chat file
        with open(labeled_fname, "wt") as fh:
            fh.write(text)



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import os
from .__about__ import __version__

BOT_CALLSIGN = os.environ.get("APRS_ASSISTANT_CALLSIGN", "<NOCALL>")

DATA_DIR = os.path.join(os.getcwd(), "data")
CACHE_DIR = os.path.join(DATA_DIR, "cache")
CHATS_DIR = os.path.join(DATA_DIR, "chats")
LABELED_DIR = os.path.join(DATA_DIR, "labeled")

USER_AGENT = "aprs-assistant/" + __version__

# Number of seconds (for cacheing etc.)
SECONDS_IN_MINUTE = 60
SECONDS_IN_HOUR = SECONDS_IN_MINUTE * 60
SECONDS_IN_DAY = SECONDS_IN_HOUR * 24
SECONDS_IN_WEEK = SECONDS_IN_DAY * 7

SESSION_TIMEOUT = 30 * SECONDS_IN_MINUTE



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import sys
import os
import json
from ._bot import generate_reply
from ._constants import BOT_CALLSIGN

if os.environ.get("OPENAI_API_KEY", "").strip() == "":
    raise ValueError("No OPENAI_API_KEY. Please set the OPENAI_API_KEY environment variable.")

if os.environ.get("BING_API_KEY", "").strip() == "":
    sys.stderr.write("Warning: No BING_API_KEY. Web searches are disabled.\nPlease set the BING_API_KEY environment variable.\n\n")

if os.environ.get("APRSFI_API_KEY", "").strip() == "":
    sys.stderr.write("Warning: No APRSFI_API_KEY. Location lookups, and localized features (e.g., local weather) are disabled.\nPlease set the APRSFI_API_KEY environment variable.\n\n")
    
if len(sys.argv) < 2:
    sys.stderr.write("SYNTAX: python -m aprs-assistant <YOUR_CALLSIGN>\n\n")
    sys.exit(0)
    
fromcall = sys.argv[1]
while True:
    request = input(f"{fromcall}: ").strip()
    if len(request) == 0:
        continue
    if request == "quit" or request == "exit":
        break
    response = generate_reply(fromcall, request)
    print(f"\n{BOT_CALLSIGN}: {response}\n")



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import os
import json
from openai import AzureOpenAI, OpenAI
#from azure.identity import DefaultAzureCredential, get_bearer_token_provider

_oai_client = None

def gpt(messages, model="gpt-4o-2024-08-06", json_mode=False, **kwargs):
    global _oai_client

    if _oai_client is None:
        _oai_client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])

    kwargs["model"] = model
    kwargs["messages"] = messages

    if json_mode == False:
        response = _oai_client.chat.completions.create(**kwargs)
        return response.choices[0].message
    else:
        kwargs["response_format"] = {"type": "json_object"}
        response = _oai_client.chat.completions.create(**kwargs)
        return json.loads(response.choices[0].message)



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import requests
import json
import os
import maidenhead

from ._cache import read_cache, write_cache
from ._constants import SECONDS_IN_MINUTE, SECONDS_IN_WEEK

def get_position(callsign):
    aprsfi_data = aprsfi_get_position(callsign)

    if aprsfi_data is None:
        return None

    lat = float(aprsfi_data["lat"])
    lon = float(aprsfi_data["lng"])

    location_data = reverse_location(lat, lon)

    result = { 
        "latitude": lat,
        "longitude": lon,
        "maidenhead_gridsquare": maidenhead.to_maiden(lat, lon, 4)
    }

    if "speed" in aprsfi_data:
        result["speed_in_kph"] = int(aprsfi_data["speed"])
    if "altitude" in aprsfi_data:
        result["altitude_in_meters"] = int(aprsfi_data["altitude"])
    if "course" in aprsfi_data:
        result["heading_in_degrees"] = int(aprsfi_data["course"])

    if "name" in location_data and len(location_data["name"]) > 0:
        result["name"] = location_data["name"]
    if "display_name" in location_data and len(location_data["display_name"]) > 0:
        result["description"] = location_data["display_name"]
    if "address" in location_data:
        result["address"] = location_data["address"]

    types = list()
    if "category" in location_data:
        types.append(location_data["category"])
    if "address_type" in location_data:
        types.append(location_data["address_type"])
    if "type" in location_data:
        types.append(location_data["type"])
    types = list(set(types))

    if len(types) > 0:
        result["category"] = ", ".join(types)

    return result


def aprsfi_get_position(callsign):
    cache_key = f"aprsfi_get_position:{callsign}"
    cached_data = read_cache(cache_key)
    if cached_data is not None:
        return cached_data
    else:
        data = _aprsfi_get_position(callsign)
        write_cache(cache_key, data, expires_in=SECONDS_IN_MINUTE*5)
        return data


def _aprsfi_get_position(callsign):
    api_key = os.environ.get("APRSFI_API_KEY", "").strip()
    if api_key == "":
        return None
    
    headers = { "User-Agent": "aprsd_gpt_plugin" }
    response = requests.get(
        f"https://api.aprs.fi/api/get?name={callsign}&what=loc&apikey={api_key}&format=json",
        headers=headers,
        stream=False
    )
    response.raise_for_status()
    response_data = response.json()

    if response_data.get("result") == "ok" and len(response_data["entries"]) > 0:
        return response_data["entries"][0]

    return None


def reverse_location(lat, lon):
    cache_key = f"reverse_location:{lat}:{lon}"
    cached_data = read_cache(cache_key)
    if cached_data is not None:
        return cached_data
    else:
        data = _reverse_location(lat, lon)
        write_cache(cache_key, data, expires_in=2*SECONDS_IN_WEEK)
        return data


def _reverse_location(lat, lon):
    headers = { "User-Agent": "aprsd_gpt_plugin" }
    response = requests.get(
        f"https://nominatim.openstreetmap.org/reverse?lat={lat}&lon={lon}&format=jsonv2",
        headers=headers,
        stream=False
    )
    response.raise_for_status()
    return response.json()



# SPDX-FileCopyrightText: 2024-present Adam Fourney <adam.fourney@gmail.com>
#
# SPDX-License-Identifier: MIT
import os
import hashlib
import json
import time

from ._constants import CACHE_DIR, SECONDS_IN_WEEK

def read_cache(key):
    hkey = _hash_key(key)
    file_path = os.path.join(CACHE_DIR, hkey + ".json")

    if not os.path.isfile(file_path):
        print(f"Cache miss: {file_path}")
        return None

    with open(file_path, "rt") as fh:
        file_data = json.loads(fh.read())
        if time.time() > file_data["expires"]:
            print(f"Cache expired: {file_path}")
            return None
        else:
            print(f"Cache hit: {file_path}")
            return file_data["data"]


def write_cache(key, data, expires_in=2*SECONDS_IN_WEEK):
    os.makedirs(CACHE_DIR, exist_ok=True)

    hkey = _hash_key(key)
    file_path = os.path.join(CACHE_DIR, hkey + ".json")

    with open(file_path, "wt") as fh:
        fh.write(json.dumps({ "expires": time.time() + expires_in, "key_hash": hkey, "key": key, "data": data }, indent=4))


def _hash_key(key):
    return hashlib.md5(json.dumps(key).encode()).hexdigest().lower()
